# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2022 CERN.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Search engine API."""

import hashlib

from flask import current_app, request

from .engine import dsl
from .proxies import current_search_client
from .utils import build_alias_name


class DefaultFilter(object):
    """Shortcut for defining default filters with query parser."""

    def __init__(self, query=None, query_parser=None):
        """Build filter property with query parser."""
        self._query = query
        self.query_parser = query_parser or (lambda x: x)

    @property
    def query(self):
        """Build lazy query if needed."""
        return self._query() if callable(self._query) else self._query

    def __get__(self, obj, objtype):
        """Return parsed query."""
        return self.query_parser(self.query)


class MinShouldMatch(str):
    """Work-around for Elasticsearch DSL problem.

    The ElasticSearch DSL Bool query tries to inspect the
    ``minimum_should_match`` parameter, but understands only integers and not
    queries like "0<1". This class circumvents the specific problematic clause
    in Elasticsearch DSL.
    """

    def __lt__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False

    def __le__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False

    def __gt__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False

    def __ge__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False


class BaseRecordsSearch(dsl.Search):
    """Example subclass for searching records using Elastic DSL."""

    class Meta:
        """Configuration for ``Search`` and ``FacetedSearch`` classes."""

        index = "*"
        fields = ("*",)
        facets = {}

        default_filter = None
        """Default filter added to search body.

        Example: ``default_filter = DefaultFilter('_access.owner:"1"')``.
        """

    def __init__(self, **kwargs):
        """Use Meta to set kwargs defaults."""
        kwargs.setdefault("index", getattr(self.Meta, "index", None))
        kwargs.setdefault("using", current_search_client)
        kwargs.setdefault("extra", {})

        min_score = current_app.config.get("SEARCH_RESULTS_MIN_SCORE")
        if min_score:
            kwargs["extra"].update(min_score=min_score)

        super(BaseRecordsSearch, self).__init__(**kwargs)

        default_filter = getattr(self.Meta, "default_filter", None)
        if default_filter:
            # NOTE: https://github.com/elastic/elasticsearch/issues/21844
            self.query = dsl.query.Bool(
                minimum_should_match=MinShouldMatch("0<1"), filter=default_filter
            )

    def get_record(self, id_):
        """Return a record by its identifier.

        :param id_: The record identifier.
        :returns: The record.
        """
        return self.query(dsl.query.Ids(values=[str(id_)]))

    def get_records(self, ids):
        """Return records by their identifiers.

        :param ids: A list of record identifier.
        :returns: A list of records.
        """
        return self.query(dsl.query.Ids(values=[str(id_) for id_ in ids]))

    @classmethod
    def faceted_search(cls, query=None, filters=None, search=None):
        """Return faceted search instance with defaults set.

        :param query: Elastic DSL query object (``Q``).
        :param filters: Dictionary with selected facet values.
        :param search: An instance of ``Search`` class. (default: ``cls()``).
        """
        search_ = search or cls()

        class RecordsFacetedSearch(dsl.FacetedSearch):
            """Pass defaults from ``cls.Meta`` object."""

            index = build_alias_name(search_._index[0])
            fields = getattr(search_.Meta, "fields", ("*",))
            facets = getattr(search_.Meta, "facets", {})

            def search(self):
                """Use ``search`` or ``cls()`` instead of default Search."""
                return search_.response_class(dsl.FacetedResponse)

        return RecordsFacetedSearch(query=query, filters=filters or {})

    def with_preference_param(self):
        """Add the preference param to the ES request and return a new Search.

        The preference param avoids the bouncing effect with multiple
        replicas, documented on ES documentation.
        See: https://www.elastic.co/guide/en/elasticsearch/guide/current
        /_search_options.html#_preference for more information.
        """
        user_hash = self._get_user_hash()
        if user_hash:
            return self.params(preference=user_hash)
        return self

    def _get_user_agent(self):
        """Retrieve the request's User-Agent, if available.

        Taken from Flask Login utils.py.
        """
        user_agent = request.headers.get("User-Agent")
        if user_agent:
            user_agent = user_agent.encode("utf-8")
        return user_agent or ""

    def _get_user_hash(self):
        """Calculate a digest based on request's User-Agent and IP address."""
        if request:
            user_hash = "{ip}-{ua}".format(
                ip=request.remote_addr, ua=self._get_user_agent()
            )
            alg = hashlib.md5()
            alg.update(user_hash.encode("utf8"))
            return alg.hexdigest()
        return None


class PrefixedIndexList(list):
    """Custom list type for avoiding double prefixing."""

    pass


class PrefixedSearchMixin:
    """Mixing to use index prefixing."""

    def prefix_index(self, index):
        """Using PrefixedIndexList type to avoid double prefixing."""
        # at object instantiation, kwargs['index'] is not defined.
        # Elasticsearch-dsl-py re-instantiated the object at each search
        # by cloning it and passing as kwargs the list of indices
        # kwargs['index'] = ['index-name1', 'index-name2']
        _index_param = index
        if not isinstance(index, PrefixedIndexList):
            if isinstance(index, (tuple, list)):
                _prefixed_index_list = [build_alias_name(_index) for _index in index]
                index = _prefixed_index_list
            elif isinstance(index, str):
                _splitted_index = index.strip().split(",")
                if len(_splitted_index) > 1:
                    _prefix_index_list = [
                        build_alias_name(_index) for _index in _splitted_index
                    ]
                    index = ",".join(_prefix_index_list)
                else:
                    index = build_alias_name(index)
                _index_param = [_index_param]
            self._original_index = _index_param

        return index

    def _clone(self):
        """Clone `_original_index` attribute.

        During re-instantiation Elasticsearch-dsl-py calls `self._clone`
        to copy over the search object. We override the method so we can
        copy the `_original_index` attribute.
        """
        s = super(PrefixedSearchMixin, self)._clone()
        s._original_index = self._original_index
        return s


class BaseRecordsSearchV2(dsl.Search):
    """Base records search V2.

    Apply configuration via kwargs instead of Meta class as in BaseRecordsSearch.
    """

    def __init__(self, fields=("*",), default_filter=None, **kwargs):
        """Sets the needed args in kwargs for the search."""
        kwargs.setdefault("index", "*")
        kwargs.setdefault("using", current_search_client)
        kwargs.setdefault("extra", {})

        min_score = current_app.config.get("SEARCH_RESULTS_MIN_SCORE")
        if min_score:
            kwargs["extra"].update(min_score=min_score)

        super(BaseRecordsSearchV2, self).__init__(**kwargs)

        if default_filter:
            # NOTE: https://github.com/elastic/elasticsearch/issues/21844
            self.query = dsl.query.Bool(
                minimum_should_match=MinShouldMatch("0<1"), filter=default_filter
            )

    def get_record(self, id_):
        """Return a record by its identifier.

        :param id_: The record identifier.
        :returns: The record.
        """
        return self.query(dsl.query.Ids(values=[str(id_)]))

    def get_records(self, ids):
        """Return records by their identifiers.

        :param ids: A list of record identifier.
        :returns: A list of records.
        """
        return self.query(dsl.query.Ids(values=[str(id_) for id_ in ids]))

    def with_preference_param(self, preference=None):
        """Add the preference param to the ES request and return a new Search.

        The preference param avoids the bouncing effect with multiple
        replicas, documented on ES documentation.
        See: https://www.elastic.co/guide/en/elasticsearch/guide/current
        /_search_options.html#_preference for more information.

        :param str preference: A preference value.
        """
        if preference:
            return self.params(preference=preference)
        return self


class RecordsSearch(PrefixedSearchMixin, BaseRecordsSearch):
    """Prefixed record search class."""

    def __init__(self, **kwargs):
        """Constructor."""
        _index = self.prefix_index(
            index=kwargs.get("index", getattr(self.Meta, "index", None))
        )
        kwargs.update({"index": _index})

        super(RecordsSearch, self).__init__(**kwargs)
        if self._index:
            self._index = PrefixedIndexList(self._index)


class RecordsSearchV2(PrefixedSearchMixin, BaseRecordsSearchV2):
    """Prefixed record search class.

    Enhanced version of RecordsSearch class to be able to inject configuration
    via kwargs, instead of using a Meta class.
    """

    def __init__(self, **kwargs):
        """Constructor."""
        _index = self.prefix_index(index=kwargs.get("index", "*"))
        kwargs.update({"index": _index})

        super(RecordsSearchV2, self).__init__(**kwargs)
        if self._index:
            self._index = PrefixedIndexList(self._index)


UnPrefixedRecordsSearch = BaseRecordsSearch
UnPrefixedRecordsSearchV2 = BaseRecordsSearchV2



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C)      2022 University Münster.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Transparency module for importing the chosen search engine.

This module imports the ``elasticsearch``/``opensearchpy`` and
``elasticsearch_dsl``/``opensearch_dsl`` packages based on availability,
and provides them as ``search`` and ``dsl`` respectively.
The aim of this is to make the choice of the search engine (Elasticsearch
vs. OpenSearch) more transparent.
"""

ES = "Elasticsearch"
OS = "OpenSearch"

try:
    # fail if both ES and OS packages are installed
    import elasticsearch
    import elasticsearch_dsl
    import opensearch_dsl
    import opensearchpy
except ModuleNotFoundError:
    # only one or zero are installed

    try:
        import elasticsearch as search
        import elasticsearch_dsl as dsl

        SearchEngine = search.Elasticsearch
        SEARCH_DISTRIBUTION = ES

    except ModuleNotFoundError:
        import opensearch_dsl as dsl
        import opensearchpy as search

        SearchEngine = search.OpenSearch
        SEARCH_DISTRIBUTION = OS

else:
    # no exception raised, both are installed. Fail.
    raise ImportError(
        "Elasticsearch and OpenSearch libraries cannot be installed both at the same time. Please uninstall the one that you are not using."
    )


def check_search_version(distribution, version):
    """Check if the search in use matches the given distribution and version.

    The ``version`` argument can either be a number or a function accepting one
    argument.
    In the first case, the specified number will be compared for equality with
    the major part of the search version (e.g. 7 for ES 7.x).
    For the second variant, the specified function will be called with the major
    version as argument. This allows for more elaborate and custom comparisons.
    The function is expected to return a boolean value.
    An example would be: ``lambda v: v >= 7``
    """
    if SEARCH_DISTRIBUTION.lower() != distribution.lower():
        return False

    if not callable(version):
        return search.VERSION[0] == version
    else:
        return version(search.VERSION[0])


def check_es_version(version):
    """Convenience alias for ``check_search_version(ES, version)``."""
    return check_search_version(ES, version)


def check_os_version(version):
    """Convenience alias for ``check_search_version(OS, version)``."""
    return check_search_version(OS, version)


def uses_es7():
    """Check if ES7+ (or OS1) is in use."""
    is_es7 = check_search_version(ES, version=lambda v: v >= 7)
    return is_es7 or check_search_version(OS, 1)


__all__ = (
    "ES",
    "OS",
    "SEARCH_DISTRIBUTION",
    "SearchEngine",
    "check_es_version",
    "check_os_version",
    "check_search_version",
    "dsl",
    "search",
    "uses_es7",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2024 CERN.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for information retrieval."""

import json
import os
import warnings

import dictdiffer
from pkg_resources import (
    iter_entry_points,
    resource_filename,
    resource_isdir,
    resource_listdir,
)
from werkzeug.utils import cached_property

from . import config
from .cli import index as index_cmd
from .engine import ES, OS, SEARCH_DISTRIBUTION, SearchEngine, dsl, search
from .errors import IndexAlreadyExistsError, NotAllowedMappingUpdate
from .utils import (
    build_alias_name,
    build_index_from_parts,
    build_index_name,
    timestamp_suffix,
)


class _SearchState(object):
    """Store connection to elastic client and registered indexes."""

    def __init__(
        self,
        app,
        entry_point_group_mappings=None,
        entry_point_group_templates=None,
        entry_point_group_component_templates=None,
        entry_point_group_index_templates=None,
        **kwargs,
    ):
        """Initialize state.

        :param app: An instance of :class:`~flask.app.Flask`.
        :param entry_point_group_mappings:
            The entrypoint group name to load mappings.
        :param entry_point_group_templates:
            The entrypoint group name to load templates.
        """
        self.app = app
        self.mappings = {}
        self.aliases = {}
        self._client = kwargs.get("client")
        self.entry_point_group_templates = entry_point_group_templates
        self.entry_point_group_component_templates = (
            entry_point_group_component_templates
        )
        self.entry_point_group_index_templates = entry_point_group_index_templates
        self._current_suffix = None

        if entry_point_group_mappings:
            self.load_entry_point_group_mappings(entry_point_group_mappings)

    @property
    def current_suffix(self):
        """Return the current suffix."""
        if self._current_suffix is None:
            self._current_suffix = timestamp_suffix()
        return self._current_suffix

    def _collect_templates(self, entrypoint_group):
        """Load actions from an entry point group."""
        result = []

        for ep in iter_entry_points(group=entrypoint_group):
            loaded_ep = ep.load()

            if callable(loaded_ep):
                for template_dir in loaded_ep():
                    result.append(self.register_templates(template_dir))
            else:
                result.append(self.register_templates(ep.module_name))

        templates = {}
        for template in result:
            for name, path in template.items():
                templates[name] = path
        return templates

    @cached_property
    def templates(self):
        """Generate a dictionary with template names and file paths."""
        return self._collect_templates(self.entry_point_group_templates)

    @cached_property
    def component_templates(self):
        """Generate a dictionary with component template names and file paths."""
        return self._collect_templates(self.entry_point_group_component_templates)

    @cached_property
    def index_templates(self):
        """Generate a dictionary with index template names and file paths."""
        return self._collect_templates(self.entry_point_group_index_templates)

    @staticmethod
    def _get_mappings_module(module):
        """Resolves the module where to find search mappings/templates.

        Finds the module that contains the search mappings/templates based
        on the current installed search distribution.

        :param module: the module/package name.
        """
        search_major_version = search.VERSION[0]
        if SEARCH_DISTRIBUTION == ES:
            subfolder = "v{}".format(search_major_version)
        elif SEARCH_DISTRIBUTION == OS:
            subfolder = "os-v{}".format(search_major_version)

            # Make sure that the OpenSearch mappings are in the folder.
            # The fallback can be removed after transition to OpenSearch.
            try:
                resource_listdir(module, subfolder)
            except FileNotFoundError:
                # fallback to ES folder with a warning if `os-vx` is not found
                subfolder = "v7"
                warnings.warn(
                    "OpenSearch v{version} mappings files not found, falling back to Elasticsearch v7 mappings for module {module}. Please add the missing OpenSearch os-v{version} mappings.".format(
                        module=module,
                        version=search_major_version,
                    )
                )
        else:
            # should never happen
            raise RuntimeError(
                "Unknown search distribution {}".format(SEARCH_DISTRIBUTION)
            )

        return "{}.{}".format(module, subfolder)

    def register_mappings(self, alias, package_name):
        """Register mappings from a package under given alias.

        :param alias: The alias.
        :param package_name: The package name.
        """
        package_name = self._get_mappings_module(package_name)

        def _walk_dir(aliases, *parts):
            root_name = build_index_from_parts(*parts)
            resource_name = os.path.join(*parts)

            data = aliases.get(root_name, {})

            for filename in resource_listdir(package_name, resource_name):
                file_path = os.path.join(resource_name, filename)

                if resource_isdir(package_name, file_path):
                    _walk_dir(data, *(parts + (filename,)))
                    continue

                filename_root, ext = os.path.splitext(filename)
                if ext not in {
                    ".json",
                }:
                    continue

                index_name = build_index_from_parts(*(parts + (filename_root,)))
                assert index_name not in data, "Duplicate index"
                filename = resource_filename(
                    package_name, os.path.join(resource_name, filename)
                )
                data[index_name] = filename
                self.mappings[index_name] = filename

            aliases[root_name] = data

        # Start the recursion here:
        _walk_dir(self.aliases, alias)

    def register_templates(self, module):
        """Register templates from the provided module.

        :param module: The templates module.
        """
        module = self._get_mappings_module(module)
        result = {}

        def _walk_dir(*parts):
            parts = parts or tuple()
            resource_name = os.path.join(*parts) if parts else ""
            for filename in resource_listdir(module, resource_name):
                file_path = os.path.join(resource_name, filename)

                if resource_isdir(module, file_path):
                    _walk_dir(*(parts + (filename,)))
                    continue

                filename_root, ext = os.path.splitext(filename)
                if ext not in {
                    ".json",
                }:
                    continue

                template_name = build_index_from_parts(*(parts + (filename_root,)))
                result[template_name] = resource_filename(
                    module, os.path.join(resource_name, filename)
                )

        # Start the recursion here:
        _walk_dir()
        return result

    def load_entry_point_group_mappings(self, entry_point_group_mappings):
        """Load actions from an entry point group."""
        for ep in iter_entry_points(group=entry_point_group_mappings):
            self.register_mappings(ep.name, ep.module_name)

    def _client_builder(self):
        """Build search engine (ES/OS) client."""
        client_config = self.app.config.get("SEARCH_CLIENT_CONFIG") or {}

        hosts = self.app.config.get("SEARCH_HOSTS")
        elastic_hosts = self.app.config.get("SEARCH_ELASTIC_HOSTS")
        if hosts is None and elastic_hosts is not None:
            warnings.warn(
                "The variable SEARCH_ELASTIC_HOSTS will be replaced by SEARCH_HOSTS",
                DeprecationWarning,
            )

        client_config.setdefault("hosts", hosts or elastic_hosts)
        return SearchEngine(**client_config)

    @property
    def client(self):
        """Return client for current application."""
        if self._client is None:
            self._client = self._client_builder()
        return self._client

    def flush_and_refresh(self, index):
        """Flush and refresh one or more indices.

        .. warning::

           Do not call this method unless you know what you are doing. This
           method is only intended to be called during tests.
        """
        prefixed_index = build_alias_name(index, app=self.app)
        self.client.indices.flush(wait_if_ongoing=True, index=prefixed_index)
        self.client.indices.refresh(index=prefixed_index)
        self.client.cluster.health(wait_for_status="yellow", request_timeout=30)
        return True

    @property
    def cluster_version(self):
        """Get version of Elasticsearch running on the cluster."""
        versionstr = self.client.info()["version"]["number"]
        return [int(x) for x in versionstr.split(".")]

    @property
    def cluster_distribution(self):
        """Get the distribution entry (opensearch or elasticsearch) on the cluster."""
        # OpenSearch provides a "distribution" field containing "opensearch"
        # Elasticsearch doesn't seem to do that
        # (checked versions: 7.10.2, 7.11.2, 7.17.5, 8.3.1)
        return self.client.info()["version"].get("distribution", "elasticsearch")

    @property
    def active_aliases(self):
        """Get a filtered list of aliases based on configuration.

        Returns aliases and their mappings that are defined in the
        `SEARCH_MAPPINGS` config variable. If the `SEARCH_MAPPINGS` is set to
        `None` (the default), all aliases are included.
        """
        whitelisted_aliases = self.app.config.get("SEARCH_MAPPINGS")
        if whitelisted_aliases is None:
            return self.aliases
        else:
            return {k: v for k, v in self.aliases.items() if k in whitelisted_aliases}

    def _get_indices(self, tree_or_filename):
        for name, value in tree_or_filename.items():
            if isinstance(value, dict):
                for result in self._get_indices(value):
                    yield result
            else:
                yield name

    def create_index(
        self,
        index,
        mapping_path=None,
        prefix=None,
        suffix=None,
        create_write_alias=True,
        ignore=None,
        dry_run=False,
    ):
        """Create index with a write alias."""
        mapping_path = mapping_path or self.mappings[index]

        final_alias = None
        alias_result = None
        # To prevent index init --force from creating a suffixed
        # index if the current instance is running without suffixes
        # make sure there is no index with the same name as the
        # alias name (i.e. the index name without the suffix).
        with open(mapping_path, "r") as body:
            final_index = build_index_name(
                index, prefix=prefix, suffix=suffix, app=self.app
            )
            if create_write_alias:
                final_alias = build_alias_name(index, prefix=prefix, app=self.app)
            index_result = (
                self.client.indices.create(
                    index=final_index,
                    body=json.load(body),
                    ignore=ignore,
                )
                if not dry_run
                else None
            )
            if create_write_alias:
                alias_result = (
                    self.client.indices.put_alias(
                        index=final_index,
                        name=final_alias,
                        ignore=ignore,
                    )
                    if not dry_run
                    else None
                )
        return (final_index, index_result), (final_alias, alias_result)

    def create(self, ignore=None, ignore_existing=False, index_list=None):
        """Yield tuple with created index name and responses from a client."""
        ignore = ignore or []
        new_indices = {}
        actions = []
        if ignore_existing and not ignore:
            ignore = [400]
        elif ignore_existing and 400 not in ignore:
            ignore.append(400)

        def ensure_not_exists(name):
            if not ignore_existing and self.client.indices.exists(name):
                raise IndexAlreadyExistsError(
                    'index/alias with name "{}" already exists'.format(name)
                )

        def _build(tree_or_filename, alias=None):
            """Build a list of index/alias actions to perform."""
            for name, value in tree_or_filename.items():
                if isinstance(value, dict):
                    _build(value, alias=name)
                else:
                    if index_list and name not in index_list:
                        continue
                    index_result, alias_result = self.create_index(
                        name, ignore=ignore, dry_run=True
                    )
                    ensure_not_exists(index_result[0])
                    new_indices[name] = index_result[0]
                    if alias_result[0]:
                        ensure_not_exists(alias_result[0])
                        actions.append(
                            dict(
                                type="create_index", index=name, create_write_alias=True
                            )
                        )
                    else:
                        actions.append(
                            dict(
                                type="create_index",
                                index=name,
                                create_write_alias=False,
                            )
                        )
            if alias:
                alias_indices = self._get_indices(tree_or_filename)
                alias_indices = [
                    new_indices[i] for i in alias_indices if i in new_indices
                ]
                if alias_indices:
                    alias_name = build_alias_name(alias, app=self.app)
                    ensure_not_exists(alias_name)
                    actions.append(
                        dict(type="create_alias", index=alias_indices, alias=alias_name)
                    )

        _build(self.active_aliases)

        for action in actions:
            if action["type"] == "create_index":
                index_result, alias_result = self.create_index(
                    action["index"],
                    create_write_alias=action.get("create_write_alias", True),
                    ignore=ignore,
                )
                yield index_result
                if alias_result[0]:
                    yield alias_result
            elif action["type"] == "create_alias":
                yield (
                    action["alias"],
                    self.client.indices.put_alias(
                        index=action["index"],
                        name=action["alias"],
                        ignore=ignore,
                    ),
                )

    def update_mapping(self, index, check=True):
        """Update mapping of the existing index."""
        mapping_path = self.mappings[index]
        index_alias_name = build_alias_name(index)

        # get api returns only dicts
        index_dict = self.client.indices.get(index_alias_name)
        index_keys = list(index_dict.keys())

        # make sure only one index exists
        assert len(index_keys) == 1

        full_index_name = index_keys[0]

        old_mapping = index_dict[full_index_name]["mappings"]

        # need to initialise Index class to use the .put_mapping API wrapper method
        index_ = dsl.Index(full_index_name, using=self.client)

        with open(mapping_path, "r") as body:
            mapping = json.load(body)["mappings"]
            changes = list(dictdiffer.diff(old_mapping, mapping))

            # allow only additions to mappings (backwards compatibility is kept)
            if not check or all([change[0] == "add" for change in changes]):
                # raises 400 if the mapping cannot be updated
                # (f.e. type changes or index needs to be closed)
                index_.put_mapping(using=self.client, body=mapping)
            else:
                non_add_changes = [change for change in changes if change[0] != "add"]
                raise NotAllowedMappingUpdate(
                    "Only additions are allowed when updating mappings to keep backwards compatibility. "
                    f"This mapping has {len(non_add_changes)} non addition changes.\n\n"
                    f"Full list of changes: {changes}"
                )

    def _replace_prefix(self, template_path, body, enforce_prefix):
        """Replace index prefix in template request body."""
        pattern = "__SEARCH_INDEX_PREFIX__"

        prefix = self.app.config["SEARCH_INDEX_PREFIX"] or ""
        if prefix:
            message = f"You are using the prefix `{prefix}`, but the template `{template_path}` does not contain the pattern `{pattern}`."
            if enforce_prefix:
                assert pattern in body, message
            else:
                warnings.warn(message)

        return body.replace(pattern, prefix)

    def _put_template(
        self, template_name, template_file, put_function, ignore, enforce_prefix=True
    ):
        """Put template in search client.

        If enforce_prefix is set to True, and the setting INVENIO_SEARCH_PREFIX_INDEX exists, then the function will
        fail if the template does not use the prefix
        """
        ignore = ignore or []
        with open(template_file, "r") as fp:
            body = fp.read()
            replaced_body = self._replace_prefix(template_file, body, enforce_prefix)
            template_name = build_alias_name(template_name, app=self.app)
            return template_file, put_function(
                name=template_name,
                body=json.loads(replaced_body),
                ignore=ignore,
            )

    def put_templates(self, ignore=None):
        """Yield tuple with registered template and response from client."""
        for template in self.templates:
            yield self._put_template(
                template,
                self.templates[template],
                self.client.indices.put_template,
                ignore,
            )

    def put_component_templates(self, ignore=None):
        for template in self.component_templates:
            yield self._put_template(
                template,
                self.component_templates[template],
                self.client.cluster.put_component_template,
                ignore,
                enforce_prefix=False,
            )

    def put_index_templates(self, ignore=None):
        for template in self.index_templates:
            yield self._put_template(
                template,
                self.index_templates[template],
                self.client.indices.put_index_template,
                ignore,
            )

    def delete(self, ignore=None, index_list=None):
        """Yield tuple with deleted index name and responses from a client."""
        ignore = ignore or []

        def _delete(tree_or_filename, alias=None):
            """Delete indexes and aliases by walking DFS."""
            # Iterate over aliases:
            for name, value in tree_or_filename.items():
                if isinstance(value, dict):
                    for result in _delete(value, alias=name):
                        yield result
                else:
                    if index_list and name not in index_list:
                        continue
                    # Resolve values to suffixed (or not) indices
                    prefixed_index = build_alias_name(name, app=self.app)
                    lookup_response = self.client.indices.get_alias(
                        index=prefixed_index, ignore=[404]
                    )
                    if "error" in lookup_response:
                        indices_to_delete = []
                    else:
                        indices_to_delete = list(lookup_response.keys())
                    if len(indices_to_delete) == 0:
                        pass
                    elif len(indices_to_delete) == 1:
                        yield (
                            name,
                            self.client.indices.delete(
                                index=indices_to_delete[0],
                                ignore=ignore,
                            ),
                        )
                    else:
                        warnings.warn(
                            (
                                "Multiple indices found during deletion of "
                                "{name}: {indices}. Deletion was skipped for them."
                            ).format(name=name, indices=indices_to_delete)
                        )

        for result in _delete(self.active_aliases):
            yield result


class InvenioSearch(object):
    """Invenio-Search extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        self._clients = {}

        if app:
            self.init_app(app, **kwargs)

    def init_app(
        self,
        app,
        entry_point_group_mappings="invenio_search.mappings",
        entry_point_group_templates="invenio_search.templates",
        entry_point_group_component_templates="invenio_search.component_templates",
        entry_point_group_index_templates="invenio_search.index_templates",
        **kwargs,
    ):
        """Flask application initialization.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        self.init_config(app)

        app.cli.add_command(index_cmd)

        state = _SearchState(
            app,
            entry_point_group_mappings=entry_point_group_mappings,
            entry_point_group_templates=entry_point_group_templates,
            entry_point_group_component_templates=entry_point_group_component_templates,
            entry_point_group_index_templates=entry_point_group_index_templates,
            **kwargs,
        )
        self._state = app.extensions["invenio-search"] = state

    @staticmethod
    def init_config(app):
        """Initialize configuration.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        for k in dir(config):
            if k.startswith("SEARCH_"):
                app.config.setdefault(k, getattr(config, k))

    def __getattr__(self, name):
        """Proxy to state object."""
        return getattr(self._state, name, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Proxy objects for easier access to application objects."""

from flask import current_app
from werkzeug.local import LocalProxy


def _get_current_search():
    """Return current state of the search extension."""
    return current_app.extensions["invenio-search"]


def _get_current_search_client():
    """Return current search client."""
    return _get_current_search().client


current_search = LocalProxy(_get_current_search)
current_search_client = LocalProxy(_get_current_search_client)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio search errors."""


class IndexAlreadyExistsError(Exception):
    """Raised when an index or alias already exists during index creation."""


class NotAllowedMappingUpdate(Exception):
    """Raised when attempted mapping update is not allowed."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2024 CERN.
# Copyright (C)      2022 University Münster.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

r"""Search management for Invenio (for Elasticsearch and OpenSearch).

Allows retrieving records from a configurable backend (currently Elasticsearch
and OpenSearch are supported).

Initialization
--------------

To be able to retrieve information from *somewhere*, we first need to setup
this *somewhere*. So make sure you have the correct version of ES/OS
installed and running (see :ref:`installation` for supported ES/OS
versions).

For running an Elasticsearch instance we recommend using
`Docker <https://docs.docker.com/install/>`_ and the official images `provided
by Elastic <https://www.docker.elastic.co/>`_
(or the OpenSearch team `<https://hub.docker.com/r/opensearchproject/opensearch>`_):

.. code-block:: console

    $ docker run -d \
        -p 9200:9200 \
        -e "discovery.type=single-node" \
        docker.elastic.co/elasticsearch/elasticsearch-oss:7.10.2

In this case, we are using Elasticsearch v7, so make sure to install
``invenio-search`` with the appropriate extras:

.. code-block:: console

    pip install invenio-search[elasticsearch7]

.. _creating_index:

Creating index
~~~~~~~~~~~~~~

To be able to run flask CLI commands later, we first need to have a Flask app,
so create the following ``app.py`` file:

.. code-block:: python

    # app.py
    from flask import Flask
    from invenio_search import InvenioSearch

    app = Flask('myapp')
    search = InvenioSearch(app)

This will create an empty index, which is not very useful, so let's add some
mappings. To not reinvent the wheel, let's reuse the mappings from the example
application. Copy the ``examples`` directory from
`Invenio-Search <https://github.com/inveniosoftware/invenio-search>`_.
Add the following line at the end of ``app.py`` file that you just created:

.. code-block:: python

    # app.py
    ...
    search.register_mappings('demo', 'examples.data')

The above code will search the directory ``examples/data`` and load all the
mapping files it can find there for the configured search engine. You can read more
about the Elasticsearch mappings in `the official documentation
<https://www.elastic.co/guide/en/elasticsearch/guide/current/mapping-intro.html>`_.

Now we can finally create the indexes. Each file in the mappings will create a
new index and a top-level alias with the name ``demo``.

.. code-block:: console

   $ export FLASK_APP=app.py
   $ flask index init

You can verify that the indices were created correctly by performing the
following requests:

.. code-block:: console

    $ # Fetch information about the "demo" alias
    $ curl http://localhost:9200/demo
    $ # Fetch information about the "demo-default-v1.0.0" alias
    $ curl http://localhost:9200/demo-default-v1.0.0

.. note::

    In earlier versions of Invenio-Search ``demo-default-v1.0.0`` was an
    index but is now a write alias pointing to a suffixed index. Read more
    about write aliases and suffixes in the Aliases_ section.

Let's index some data. Open ``flask shell`` and index a document with the
following code:

.. code-block:: python

    import json
    from invenio_search import current_search_client

    current_search_client.index(
        index='demo-default-v1.0.0',
        body=json.dumps({
            'title': 'Hello invenio-search',
            'body': 'test 1'
        })
    )

No error message? Good! You can see that your new document was indexed by
going to http://localhost:9200/demo/_search.

Searching for data
~~~~~~~~~~~~~~~~~~

Let's try to retrieve data in a programmatic way. Start a Python REPL and run
the following commands.

First, let's initialize app from the ``app.py`` file that we created in the
previous step.
Python REPL.

.. code-block:: python

   from app import app

We will need a Flask application context, so let's push one:

.. code-block:: python

   app.app_context().push()

Create a custom search class that will search for ``example`` type of
documents inside the ``demo`` alias (or you could use the default
``RecordsSearch`` class to search for all document types in all indexes):

.. code-block:: python

   from invenio_search import RecordsSearch
   class ExampleSearch(RecordsSearch):
       class Meta:
           index = 'demo'
           fields = ('*', )
           facets = {}
   search = ExampleSearch()

Let's find all documents:

.. code-block:: python

   response = search.execute()
   response.to_dict()

If everything went well, you should now see that we have 1 hit - a document
with ``Hello invenio-search`` title. If you get the ``TransportError(404,
'index_not_found_exception', 'no such index')`` error - it means that you
forgot to create the index (follow the steps from :ref:`creating_index` to
see how to setup an index and add example data).

Creating a search page
~~~~~~~~~~~~~~~~~~~~~~

Let's create a simple web page where you can send queries to the search engine
and see the results.
Create a new app.py file with a route.

.. code-block:: python

    # app.py
    # note: 'invenio_search.engine' imports the installed search DSL dependency library
    #       and makes it available as 'dsl'
    from invenio_search.engine import dsl
    from flask import Flask, jsonify, request
    from invenio_search import InvenioSearch, RecordsSearch

    app = Flask('myapp')

    search = InvenioSearch(app)

    # This line is needed to be able to call `flask index init`
    search.register_mappings('demo', 'examples.data')


    @app.route('/', methods=['GET', 'POST'])
    def index():
        search = RecordsSearch()
        if 'q' in request.values:
            search = search.query(dsl.QueryString(query=request.values.get('q')))

        return jsonify(search.execute().to_dict())

Run example development server:

.. code-block:: console

   $ FLASK_DEBUG=1 FLASK_APP=app.py flask run -p 5000

And now you can perform search queries:

.. code-block:: console

   $ curl http://localhost:5000/?q=body:test

Filtering
~~~~~~~~~
To filter out some documents, you can create your own search class. Let's try
to remove all private documents from the search results (by ``private``
documents, we understand all the documents that have ``public`` attribute set
to 0).

Open ``flask shell`` and add one public and one private document to the search engine:

.. code-block:: python

    import json
    from invenio_search import current_search_client

    # Index public document
    current_search_client.index(
        index='demo-default-v1.0.0',
        body=json.dumps({
            'title': 'Public',
            'body': 'test 1',
            'public': 1
        })
    )
    # Index private document
    current_search_client.index(
        index='demo-default-v1.0.0',
        body=json.dumps({
            'title': 'Private',
            'body': 'test 1',
            'public': 0
        })
    )

Now, create a new search class that will return all documents of type
``example`` from the ``demo`` index and select only the public ones
(documents where ``public`` is set to 1):

.. code-block:: python

    # app.py
    from invenio_search.engine import dsl

    class PublicSearch(RecordsSearch):
        class Meta:
            index = 'demo'
            fields = ('*', )
            facets = {}

        def __init__(self, **kwargs):
            super(PublicSearch, self).__init__(**kwargs)
            self.query = dsl.Q(
                dsl.Bool(filter=[dsl.Q('term', public=1)])
            )

Update the ``index`` function and replace the search class with our new
``PublicSearch`` class:

.. code-block:: python

    # app.py
    @app.route('/', methods=['GET', 'POST'])
    def index():
        search = PublicSearch()
        ...

Now, you can search for documents with ``test`` in the body.

.. code-block:: console

   $ curl http://localhost:5000/?q=body:test

You should find only one document - the one with ``Public`` title.

This is a very simple example of how to filter out some records.
If you want to define role based access rights control,
check the invenio-access_ module.


Miscellaneous
-------------

OpenSearch/Elasticsearch version support
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Major versions of OpenSearch/Elasticsearch can include breaking changes to mappings so
mappings for each version of OpenSearch/Elasticsearch are stored in separate folders.
Invenio-Search will use these mappings when creating the indices. You can
see a full example in the ``examples/data`` directory of the Invenio-Search
repository:

.. code-block:: console

    $ tree --dirsfirst examples/data

    examples/data
    +- os-v1
    |  +- demo        # OpenSearch 1 mappings
    |  |  +- authorities
    |  |  |  +- authority-v1.0.0.json
    |  |  +- bibliographic
    |  |  |  +- bibliographic-v1.0.0.json
    |  |  +- default-v1.0.0.json
    |  +- __init__.py
    +- v7
    |  +- demo        # Elasticsearch 7 mappings
    |  |  +- authorities
    |  |  |  +- authority-v1.0.0.json
    |  |  +- bibliographic
    |  |  |  +- bibliographic-v1.0.0.json
    |  |  +- default-v1.0.0.json
    |  +- __init__.py
    +-- __init__.py

Elasticsearch plugins
~~~~~~~~~~~~~~~~~~~~~

For convenience, you can install a plugin like `Elastic HQ
<http://www.elastichq.org/>`_ for easy introspection of your indexes and their
content. Otherwise, you can use curl as described `in the official
documentation <https://www.elastic.co/guide/en/elasticsearch/guide/current/
_talking_to_elasticsearch.html>`_.

.. _invenio-access:  https://invenio-access.readthedocs.io/

.. _Aliases:

Indexes and aliases
~~~~~~~~~~~~~~~~~~~

.. graphviz::

    digraph D {
        rankdir="LR"
        node[shape=record, fixedsize=true, width=1]

        R[label="records", style=rounded]
        RD[label="records-dataset-v1.0.0", style="rounded,dashed", width=2.5]
        RP[label="records-paper-v1.0.0", style="rounded,dashed", width=2.5]
        RDI[label="records-dataset-v1.0.0-1564056972", width=4]
        RPI[label="records-paper-v1.0.0-1564056972", width=4]

        A[label="authors", style=rounded]
        AA[label="authors-author-v1.0.0", style="rounded,dashed", width=2.5]
        AAI[label="authors-author-v1.0.0-1564056972", width=4]

        R -> RDI
        R -> RPI
        RP -> RPI
        RD -> RDI

        A -> AAI
        AA -> AAI
    }

Indexes and aliases are organized as seen in the graph above. This example has
three "concrete" indexes:

- ``authors-author-v1.0.0-1564056972``
- ``records-dataset-v1.0.0-1564056972``
- ``records-paper-v1.0.0-1564056972``

They all share the suffix ``1564056972``. Each index though has also a
corresponding "write alias" with its un-suffixed name:

- ``authors-author-v1.0.0 -> authors-author-v1.0.0-1564056972``
- ``records-dataset-v1.0.0 -> records-dataset-v1.0.0-1564056972``
- ``records-paper-v1.0.0 -> records-paper-v1.0.0-1564056972``

The other aliases in the example, ``records`` and ``authors``, are top-level
aliases pointing to all the indexes in their same hierarchy:

- ``authors -> authors-author-v1.0.0-1564056972``
- ``records -> records-dataset-v1.0.0-1564056972``
- ``records -> records-paper-v1.0.0-1564056972``

Top-level **aliases** are aliases that can point to one or multiple
indexes. The purpose of these aliases is to group indexes and be able to
perform searches over multiple indexes. These aliases should never be indexed
to as the indexing will fail if they point to multiple indexes.

The other type of alias is the **write alias** which is an alias that only
points to a single index and has the same name as the index without the suffix.
This alias should be used whenever you need to index something. The name of the
write alias is the same as the un-suffixed index name to allow backwards
compatibilty with previous versions of Invenio-Search.

An **index** ends with a suffix which is the timestamp of the index creation
time. The suffix allows multiple revisions of the same index to exist at the
same time. This is useful if you want to update the mappings of an index and
migrate to a new index. With suffixes, it's possible to keep the two versions
of the same index and sync them. When the migration is completed the write
alias can be pointed to the new index and the application will use the new
index. This allows in-cluster migrations without any downtime.

More information about index migrations can be found in the
`Invenio-Index-Migrator
<https://github.com/inveniosoftware/invenio-index-migrator>`_.
"""

from .api import (
    RecordsSearch,
    RecordsSearchV2,
    UnPrefixedRecordsSearch,
    UnPrefixedRecordsSearchV2,
)
from .ext import InvenioSearch
from .proxies import current_search, current_search_client

__version__ = "2.4.2"


__all__ = (
    "__version__",
    "InvenioSearch",
    "RecordsSearch",
    "RecordsSearchV2",
    "UnPrefixedRecordsSearch",
    "UnPrefixedRecordsSearchV2",
    "current_search",
    "current_search_client",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2024 CERN.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Click command-line interface for managing search indexes."""

import json
import sys
from functools import wraps

import click
from flask.cli import with_appcontext

from .engine import SEARCH_DISTRIBUTION, search
from .proxies import current_search, current_search_client


def abort_if_false(ctx, param, value):
    """Abort command if value is False."""
    if not value:
        ctx.abort()


def search_version_check(f):
    """Decorator to check the version of the search engine."""

    @wraps(f)
    def inner(*args, **kwargs):
        # NOTE: here we want to compare the *actual* versions of the library and search
        client_ver = search.VERSION[0]
        cluster_ver = current_search.cluster_version[0]
        cluster_distro = current_search.cluster_distribution

        if SEARCH_DISTRIBUTION.lower() != cluster_distro:
            raise click.ClickException(
                "Search distribution mismatch. Invenio was installed with "
                "{expected} support, but the cluster runs {running}.".format(
                    expected=SEARCH_DISTRIBUTION.lower(),
                    running=cluster_distro,
                )
            )

        has_same_major_ver = cluster_ver == client_ver
        is_opensearch_v2_client = (
            SEARCH_DISTRIBUTION.lower() == "opensearch" and client_ver == 2
        )
        is_compatible = (
            has_same_major_ver
            # OpenSearch v2.x client is compatible with both v2.x and v3.x cluster versions
            or (is_opensearch_v2_client and cluster_ver in (2, 3)),
        )

        if not is_compatible:
            raise click.ClickException(
                "{search} version mismatch. Invenio was installed with "
                "{search} v{client_ver}.x support, but the cluster runs "
                "{search} v{cluster_ver}.x.".format(
                    search=SEARCH_DISTRIBUTION.lower(),
                    client_ver=client_ver,
                    cluster_ver=cluster_ver,
                )
            )
        return f(*args, **kwargs)

    return inner


#
# Index management commands
#
@click.group()
def index():
    """Manage search indices."""


@index.command()
@with_appcontext
@search_version_check
def check():
    """Check search engine version."""
    click.secho("Checks passed", fg="green")


@index.command()
@click.option("--force", is_flag=True, default=False)
@with_appcontext
@search_version_check
def init(force):
    """Initialize registered aliases and mappings."""
    actions = [
        {
            "title": "Putting templates",
            "function": current_search.put_templates,
            "items": current_search.templates,
        },
        {
            "title": "Putting component templates",
            "function": current_search.put_component_templates,
            "items": current_search.component_templates,
        },
        {
            "title": "Putting index templates",
            "function": current_search.put_index_templates,
            "items": current_search.index_templates,
        },
        {
            "title": "Creating indexes",
            "function": current_search.create,
            "items": current_search.mappings,
        },
    ]
    for action in actions:
        if len(action["items"]) == 0:
            click.secho(
                "There is no need for '" + action["title"] + "' (no items specified)",
                fg="green",
                bold=True,
                file=sys.stderr,
            )
            continue

        click.secho(action["title"] + "...", fg="green", bold=True, file=sys.stderr)
        with click.progressbar(
            action["function"](ignore=[400] if force else None),
            length=len(action["items"]),
        ) as bar:
            for name, response in bar:
                bar.label = name


@index.command()
@click.option(
    "--yes-i-know",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Do you know that you are going to destroy all indexes?",
)
@click.option("--force", is_flag=True, default=False)
@with_appcontext
@search_version_check
def destroy(force):
    """Destroy all indexes."""
    click.secho("Destroying indexes...", fg="red", bold=True, file=sys.stderr)
    with click.progressbar(
        current_search.delete(ignore=[400, 404] if force else None),
        length=len(current_search.mappings),
    ) as bar:
        for name, response in bar:
            bar.label = name


@index.command()
@click.argument("index_name")
@click.option("-b", "--body", type=click.File("r"), default=sys.stdin)
@click.option("--force", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@with_appcontext
@search_version_check
def create(index_name, body, force, verbose):
    """Create a new index."""
    result = current_search_client.indices.create(
        index=index_name,
        body=json.load(body),
        ignore=[400] if force else None,
    )
    if verbose:
        click.echo(json.dumps(result))


@index.command()
@click.argument("index_name")
@click.option("--check/--no-check", is_flag=True, default=True)
@with_appcontext
@search_version_check
def update(index_name, check=True):
    """Update mappings of existing index."""
    current_search.update_mapping(index_name, check=check)
    click.secho(f"Mapping of index {index_name} updated successfully.", fg="green")


@index.command("list")
@click.option("-a", "--only-active", is_flag=True, default=False)
@click.option("--only-aliases", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@with_appcontext
def list_cmd(only_active, only_aliases, verbose):
    """List indices."""

    def _tree_print(d, rec_list=None, verbose=False, indent=2):
        # Note that on every recursion rec_list is copied,
        # which might not be very effective for very deep dictionaries.
        rec_list = rec_list or []
        for idx, key in enumerate(sorted(d)):
            line = [
                "│" + " " * indent if i == 1 else " " * (indent + 1) for i in rec_list
            ]
            line.append("└──" if len(d) - 1 == idx else "├──")
            click.echo("".join(line), nl=False)
            if isinstance(d[key], dict):
                click.echo(key)
                new_rec_list = rec_list + [0 if len(d) - 1 == idx else 1]
                _tree_print(d[key], new_rec_list, verbose)
            else:
                leaf_txt = "{} -> {}".format(key, d[key]) if verbose else key
                click.echo(leaf_txt)

    aliases = current_search.active_aliases if only_active else current_search.aliases
    active_aliases = current_search.active_aliases

    if only_aliases:
        click.echo(json.dumps(list((aliases.keys())), indent=4))
    else:
        # Mark active indices for printout
        aliases = {
            (k + (" *" if k in active_aliases else "")): v for k, v in aliases.items()
        }
        click.echo(_tree_print(aliases, verbose=verbose))


@index.command()
@click.argument("index_name")
@click.option("--force", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@click.option(
    "--yes-i-know",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Do you know that you are going to delete the index?",
)
@with_appcontext
@search_version_check
def delete(index_name, force, verbose):
    """Delete index by its name."""
    result = current_search_client.indices.delete(
        index=index_name,
        ignore=[400, 404] if force else None,
    )
    if verbose:
        click.echo(json.dumps(result))


@index.command()
@click.argument("index_name")
@click.option("-i", "--identifier", default=None)
@click.option("-b", "--body", type=click.File("r"), default=sys.stdin)
@click.option("--force", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@with_appcontext
@search_version_check
def put(index_name, identifier, body, force, verbose):
    """Index input data."""
    result = current_search_client.index(
        index=index_name,
        id=identifier,
        body=json.load(body),
        op_type="index" if force or identifier is None else "create",
    )

    if verbose:
        click.echo(json.dumps(result))
